import argparse
import logging
import multiprocessing
from pathlib import Path
import sys
from datetime import date, timedelta, datetime
from typing import List

from oonidata.cli.sync import run_sync
from oonidata.cli.process import run_process

log = logging.getLogger("oonidata")

log.addHandler(logging.StreamHandler())
log.setLevel(logging.INFO)


def _parse_date_flag(date_str: str) -> date:
    return datetime.strptime(date_str, "%Y-%m-%d").date()


def _parse_csv(s: str) -> List[str]:
    return s.split(",")


def arg_test_name(p: argparse.ArgumentParser):
    p.add_argument(
        "--test-name",
        type=_parse_csv,
        help="test_name you care to process, can be comma separated for a list (eg. web_connectivity,whatsapp). If omitted will select process all test names.",
    )


def arg_start_day(p: argparse.ArgumentParser):
    p.add_argument(
        "--start-day",
        type=_parse_date_flag,
        default=date.today() - timedelta(days=14),
        help=(
            "the timestamp of the day for which we should start processing data (inclusive). "
            "Note: this is the upload date, which doesn't necessarily match the measurement date."
        ),
    )


def arg_end_day(p: argparse.ArgumentParser):
    p.add_argument(
        "--end-day",
        type=_parse_date_flag,
        default=date.today() + timedelta(days=1),
        help="the timestamp of the day for which we should stop processing data, this date is not included.",
    )


def arg_probe_cc(p: argparse.ArgumentParser):
    p.add_argument(
        "--probe-cc",
        type=_parse_csv,
        help="two letter country code, can be comma separated for a list (eg. IT,US). If omitted will select process all countries.",
    )


def main():
    parser = argparse.ArgumentParser("OONI Data tools")
    parser.set_defaults(func=lambda r: parser.print_usage())

    subparsers = parser.add_subparsers()

    # oonidata sync command
    parser_sync = subparsers.add_parser(
        "sync", help="Sync OONI measurements to a directory"
    )
    arg_probe_cc(parser_sync)
    arg_test_name(parser_sync)
    arg_start_day(parser_sync)
    arg_end_day(parser_sync)
    parser_sync.add_argument("--max-string-size", type=int)
    parser_sync.add_argument("--output-dir", type=Path, required=True)
    parser_sync.add_argument("--debug", action="store_true")
    parser_sync.set_defaults(func=run_sync)

    # oonidata process command
    parser_process = subparsers.add_parser(
        "process", help="Process OONI measurements to clickhouse or csv"
    )
    parser_process.add_argument(
        "--csv-dir",
        type=Path,
    )
    parser_process.add_argument(
        "--clickhouse",
        type=str,
    )
    parser_process.add_argument(
        "--data-dir",
        default=Path("datadir"),
        type=Path,
    )
    parser_process.add_argument(
        "--parallelism",
        type=int,
        default=multiprocessing.cpu_count(),
        help="number of processes to use. Only works when writing to a database.",
    )
    parser_process.add_argument(
        "--start-at-idx",
        type=int,
        default=0,
    )
    arg_probe_cc(parser_process)
    arg_test_name(parser_process)
    arg_start_day(parser_process)
    arg_end_day(parser_process)
    parser_process.add_argument("--only-verdicts", action="store_true")
    parser_process.add_argument("--skip-verdicts", action="store_true")
    parser_process.add_argument("--fast-fail", action="store_true")
    parser_process.set_defaults(func=run_process)

    args = parser.parse_args()
    sys.exit(args.func(args))


if __name__ == "__main__":
    main()



import logging
import multiprocessing
from oonidata.dataclient import date_interval
from oonidata.db.connections import CSVConnection, ClickhouseConnection
from oonidata.fingerprintdb import FingerprintDB
from oonidata.netinfo import NetinfoDB
from oonidata.processing import (
    generate_website_verdicts,
    process_day,
    write_verdicts_to_db,
)


log = logging.getLogger("oonidata.processing")


def worker(day_queue, args):
    fingerprintdb = FingerprintDB(datadir=args.data_dir, download=False)
    netinfodb = NetinfoDB(datadir=args.data_dir, download=False)

    if args.clickhouse:
        db = ClickhouseConnection(args.clickhouse)
    elif args.csv_dir:
        db = CSVConnection(args.csv_dir)
    else:
        raise Exception("Missing --csv-dir or --clickhouse")

    skip_verdicts = args.skip_verdicts
    if not isinstance(db, ClickhouseConnection):
        skip_verdicts = True

    while True:
        day = day_queue.get(block=True)
        if day == None:
            break

        if isinstance(db, ClickhouseConnection) and args.only_verdicts:
            write_verdicts_to_db(
                db,
                generate_website_verdicts(
                    args.day,
                    db,
                    fingerprintdb,
                    netinfodb,
                ),
            )
            continue

        process_day(
            db,
            fingerprintdb,
            netinfodb,
            day,
            test_name=args.test_name,
            probe_cc=args.probe_cc,
            start_at_idx=args.start_at_idx,
            skip_verdicts=skip_verdicts,
            fast_fail=args.fast_fail,
        )

    db.close()


def run_process(args):
    # This triggers the download of the required files
    FingerprintDB(datadir=args.data_dir, download=True)
    NetinfoDB(datadir=args.data_dir, download=True)

    if args.csv_dir:
        log.info(
            "When generating CSV outputs we currently only support parallelism of 1"
        )
        args.parallelism = 1

    day_queue = multiprocessing.Queue()
    pool = multiprocessing.Pool(
        processes=args.parallelism, initializer=worker, initargs=(day_queue, args)
    )
    for day in date_interval(args.start_day, args.end_day):
        day_queue.put(day)

    for _ in range(args.parallelism):
        day_queue.put(None)

    day_queue.close()
    day_queue.join_thread()

    pool.close()
    pool.join()



import logging
from typing import List, Optional, NamedTuple
from enum import Enum
from datetime import datetime
from dataclasses import dataclass
from oonidata.dataformat import SIGNAL_PEM_STORE
from oonidata.fingerprintdb import FingerprintDB
from oonidata.netinfo import NetinfoDB
from oonidata.datautils import (
    TLSCertStore,
    InvalidCertificateChain,
)

from oonidata.observations import (
    DNSObservation,
    HTTPObservation,
    NettestObservation,
    Observation,
    TCPObservation,
    TLSObservation,
)

log = logging.getLogger("oonidata.events")


class OutcomeType(Enum):
    # k: everything is OK
    OK = "k"
    # b: blocking is happening with an unknown scope
    BLOCKED = "b"
    # n: national level blocking
    NATIONAL_BLOCK = "n"
    # i: isp level blocking
    ISP_BLOCK = "i"
    # l: local blocking (school, office, home network)
    LOCAL_BLOCK = "l"
    # s: server-side blocking
    SERVER_SIDE_BLOCK = "s"
    # d: the subject is down
    DOWN = "d"
    # t: this is a signal indicating some form of network throttling
    THROTTLING = "t"


def fp_scope_to_outcome(scope: Optional[str]) -> OutcomeType:
    # "nat" national level blockpage
    # "isp" ISP level blockpage
    # "prod" text pattern related to a middlebox product
    # "inst" text pattern related to a voluntary instition blockpage (school, office)
    # "vbw" vague blocking word
    # "fp" fingerprint for false positives
    if scope == "nat":
        return OutcomeType.NATIONAL_BLOCK
    elif scope == "isp":
        return OutcomeType.ISP_BLOCK
    elif scope == "inst":
        return OutcomeType.LOCAL_BLOCK
    elif scope == "fp":
        return OutcomeType.SERVER_SIDE_BLOCK
    return OutcomeType.BLOCKED


class Outcome(NamedTuple):
    outcome_type: OutcomeType
    outcome_subject: str
    outcome_detail: str
    outcome_meta: dict
    confidence: float


@dataclass
class BlockingEvent:
    measurement_uid: str
    report_id: str
    input: str
    timestamp: datetime

    probe_asn: int
    probe_cc: str

    probe_as_org_name: str
    probe_as_cc: str

    network_type: str

    resolver_ip: Optional[str]
    resolver_asn: Optional[int]
    resolver_as_org_name: Optional[str]
    resolver_as_cc: Optional[str]
    resolver_cc: Optional[str]

    observation_ids: List[str]
    outcomes: List[Outcome]
    ok_confidence: float

    anomaly: bool
    confirmed: bool


@dataclass
class WebsiteBlockingEvent(BlockingEvent):
    domain_name: str
    website_name: str


class SignalBlockingEvent(BlockingEvent):
    pass


def make_base_event_meta(obs: Observation) -> dict:
    return dict(
        measurement_uid=obs.measurement_uid,
        report_id=obs.report_id,
        input=obs.input,
        timestamp=obs.timestamp,
        probe_asn=obs.probe_asn,
        probe_cc=obs.probe_cc,
        probe_as_org_name=obs.probe_as_org_name,
        probe_as_cc=obs.probe_as_cc,
        network_type=obs.network_type,
        resolver_ip=obs.resolver_ip,
        resolver_asn=obs.resolver_asn,
        resolver_as_org_name=obs.resolver_as_org_name,
        resolver_as_cc=obs.resolver_as_cc,
        resolver_cc=obs.resolver_cc,
    )


def make_signal_blocking_event(
    nt_obs: NettestObservation,
    dns_o_list: List[DNSObservation],
    tcp_o_list: List[TCPObservation],
    tls_o_list: List[TLSObservation],
    http_o_list: List[HTTPObservation],
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
) -> SignalBlockingEvent:
    # We got nothin', we can't do nothin'
    assert len(dns_o_list) > 0, "empty DNS observation list"

    confirmed = False
    anomaly = False

    signal_is_down = False

    outcomes = []
    observation_ids = []
    inconsistent_dns_answers = set()

    for dns_o in dns_o_list:
        observation_ids.append(dns_o.observation_id)
        if dns_o.domain_name == "uptime.signal.org":
            # This DNS query is used by signal to figure out if some of it's
            # services are down.
            # see: https://github.com/signalapp/Signal-Android/blob/c4bc2162f23e0fd6bc25941af8fb7454d91a4a35/app/src/main/java/org/thoughtcrime/securesms/jobs/ServiceOutageDetectionJob.java#L25
            # TODO: should we do something in the case in which we can't tell
            # because DNS blocking is going on (ex. in Iran)?
            if dns_o.answer == "127.0.0.2":
                signal_is_down = True
            continue

        if dns_o.failure:
            anomaly = True
            outcomes.append(
                Outcome(
                    outcome_type=OutcomeType.BLOCKED,
                    outcome_subject=f"{dns_o.domain_name}",
                    outcome_detail=f"dns.{dns_o.failure}",
                    outcome_meta={},
                    confidence=0.8,
                )
            )
            continue

        if not dns_o.is_tls_consistent:
            # We don't set the anomaly flag, because this logic is very
            # susceptible to false positives
            # anomaly = True
            outcome_meta = {"why": "tls_inconsistent", "ip": dns_o.answer}
            outcome_type = OutcomeType.BLOCKED
            fp = fingerprintdb.match_dns(dns_o.answer)
            if fp:
                outcome_type = fp_scope_to_outcome(fp.scope)
                if outcome_type != OutcomeType.SERVER_SIDE_BLOCK:
                    confirmed = True
                    anomaly = True
                outcome_meta["fp_name"] = fp.name

            confidence = 0.3
            # Having a TLS inconsistency is a much stronger indication than not
            # knowing.
            if dns_o.is_tls_consistent == False:
                inconsistent_dns_answers.add(dns_o.answer)
                # Only set the anomaly here
                anomaly = True
                confidence = 0.8

            outcomes.append(
                Outcome(
                    outcome_type=outcome_type,
                    outcome_detail="dns.inconsistent",
                    outcome_subject=f"{dns_o.domain_name}",
                    outcome_meta=outcome_meta,
                    confidence=confidence,
                )
            )
            continue

    if signal_is_down:
        # The service is down. No point in going on with the analysis
        # It's still possible for the service to be down, yet there to be DNS
        # level interference, so we still count the other outcomes if they were
        # present.
        outcomes.append(
            Outcome(
                outcome_type=OutcomeType.DOWN,
                outcome_subject="Signal Messenger",
                outcome_detail="down",
                outcome_meta={},
                confidence=0.9,
            )
        )
        return SignalBlockingEvent(
            outcomes=outcomes,
            observation_ids=observation_ids,
            anomaly=anomaly,
            confirmed=confirmed,
            **make_base_event_meta(nt_obs),
        )

    for tcp_o in tcp_o_list:
        observation_ids.append(tcp_o.observation_id)
        if tcp_o.failure and tcp_o.ip not in inconsistent_dns_answers:
            anomaly = True
            outcomes.append(
                Outcome(
                    outcome_type=OutcomeType.BLOCKED,
                    outcome_subject=f"{tcp_o.ip}:{tcp_o.port}",
                    outcome_detail=f"tcp.{tcp_o.failure}",
                    outcome_meta={},
                    confidence=0.7,
                )
            )

    cert_store = TLSCertStore(pem_cert_store=SIGNAL_PEM_STORE)
    for tls_o in tls_o_list:
        observation_ids.append(tls_o.observation_id)
        # We skip analyzing TLS handshakes that are the result of an
        # inconsistent DNS resolution.
        if tls_o.ip in inconsistent_dns_answers:
            continue

        if tls_o.failure and not tls_o.failure.startswith("ssl_"):
            anomaly = True
            outcomes.append(
                Outcome(
                    outcome_type=OutcomeType.BLOCKED,
                    outcome_subject=f"{tls_o.server_name}",
                    outcome_detail=f"tls.{tls_o.failure}",
                    outcome_meta={},
                    confidence=0.7,
                )
            )

        if tls_o.peer_certificates:
            try:
                _, _ = cert_store.validate_cert_chain(
                    tls_o.timestamp, tls_o.peer_certificates
                )
                # The server_name is listed in the SAN only for the older certs.
                # Since we are pinning to only the two known signal CAs it's
                # probably safe to just ignore.
            except InvalidCertificateChain as exc:
                anomaly = True
                outcomes.append(
                    Outcome(
                        outcome_type=OutcomeType.BLOCKED,
                        outcome_subject=f"{tls_o.server_name}",
                        outcome_detail=f"tls.ssl_invalid_certificate",
                        outcome_meta={"cert_error": str(exc)},
                        confidence=0.9,
                    )
                )

    # This is an upper bound, which means we might be over-estimating blocking
    ok_confidence = 1 - max(map(lambda o: o.confidence, outcomes), default=0)
    outcomes.append(
        Outcome(
            outcome_type=OutcomeType.OK,
            outcome_subject=f"Signal Messenger",
            outcome_detail=f"ok",
            outcome_meta={},
            confidence=ok_confidence,
        )
    )

    # TODO: add support for validating if the HTTP responses are also consistent
    return SignalBlockingEvent(
        outcomes=outcomes,
        observation_ids=observation_ids,
        anomaly=anomaly,
        confirmed=confirmed,
        ok_confidence=ok_confidence,
        **make_base_event_meta(nt_obs),
    )



import time
import logging
import traceback
import orjson

from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from datetime import datetime, date, timedelta
from pathlib import Path
from dataclasses import asdict, fields

from typing import Tuple, List, Generator, Type, TypeVar, Optional, Union, Iterable

from oonidata.datautils import one_day_dict, is_ip_bogon
from oonidata.observations import (
    NettestObservation,
    DNSObservation,
    HTTPObservation,
    Observation,
    TCPObservation,
    TLSObservation,
    make_http_observations,
    make_dns_observations,
    make_tcp_observations,
    make_tls_observations,
    make_web_connectivity_observations,
    make_ip_to_domain,
)
from oonidata.dataformat import DNSCheck, load_measurement
from oonidata.dataformat import BaseMeasurement, WebConnectivity, Tor
from oonidata.fingerprintdb import FingerprintDB
from oonidata.netinfo import NetinfoDB
from oonidata.verdicts import (
    DNSBaseline,
    Verdict,
    make_dns_baseline,
    make_http_baseline_map,
    make_tcp_baseline_map,
    make_website_verdicts,
)

from oonidata.dataclient import (
    MeasurementListProgress,
    iter_measurements,
    ProgressStatus,
)
from oonidata.db.connections import (
    DatabaseConnection,
    ClickhouseConnection,
    CSVConnection,
)

log = logging.getLogger("oonidata.processing")


def observation_field_names(obs_class: Type[Observation]) -> List[str]:
    return list(map(lambda dc: dc.name, fields(obs_class)))


def make_observation_row(observation: Observation) -> dict:
    return asdict(observation)


def make_verdict_row(v: Verdict) -> dict:
    row = asdict(v)
    # XXX come up with a cleaner solution to this
    row["outcome"] = row["outcome"].value
    return row


def write_observations_to_db(
    db: DatabaseConnection, observations: Iterable[Observation]
) -> None:
    for obs in observations:
        row = make_observation_row(obs)
        db.write_row(obs.__table_name__, row)


def write_verdicts_to_db(db: DatabaseConnection, verdicts: Iterable[Verdict]) -> None:
    for v in verdicts:
        row = make_verdict_row(v)
        db.write_row("verdict", row)


def default_processor(
    msmt: BaseMeasurement,
    db: DatabaseConnection,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
) -> None:
    print(f"Ignoring {msmt}")


def tor_processor(
    msmt: Tor,
    db: DatabaseConnection,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
) -> None:

    ip_to_domain = {}
    for target_id, target_msmt in msmt.test_keys.targets.items():
        write_observations_to_db(
            db,
            make_http_observations(
                msmt, target_msmt.requests, fingerprintdb, netinfodb, target=target_id
            ),
        )

        write_observations_to_db(
            db,
            make_dns_observations(
                msmt, target_msmt.queries, fingerprintdb, netinfodb, target=target_id
            ),
        )

        write_observations_to_db(
            db,
            make_tcp_observations(
                msmt, target_msmt.tcp_connect, netinfodb, ip_to_domain, target=target_id
            ),
        )

        write_observations_to_db(
            db,
            make_tls_observations(
                msmt,
                target_msmt.tls_handshakes,
                target_msmt.network_events,
                netinfodb,
                ip_to_domain,
            ),
        )


def web_connectivity_processor(
    msmt: WebConnectivity,
    db: DatabaseConnection,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
) -> None:
    write_observations_to_db(
        db,
        make_web_connectivity_observations(
            msmt, fingerprintdb=fingerprintdb, netinfodb=netinfodb
        ),
    )


def dnscheck_processor(
    msmt: DNSCheck,
    db: DatabaseConnection,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
) -> None:
    ip_to_domain = {}
    if msmt.test_keys.bootstrap:
        dns_observations = list(
            make_dns_observations(
                msmt, msmt.test_keys.bootstrap.queries, fingerprintdb, netinfodb
            )
        )
        ip_to_domain = make_ip_to_domain(dns_observations)
        write_observations_to_db(
            db,
            dns_observations,
        )

    lookup_map = msmt.test_keys.lookups or {}
    for lookup in lookup_map.values():
        write_observations_to_db(
            db, make_dns_observations(msmt, lookup.queries, fingerprintdb, netinfodb)
        )

        write_observations_to_db(
            db,
            make_http_observations(msmt, lookup.requests, fingerprintdb, netinfodb),
        )

        write_observations_to_db(
            db,
            make_tcp_observations(msmt, lookup.tcp_connect, netinfodb, ip_to_domain),
        )

        write_observations_to_db(
            db,
            make_tls_observations(
                msmt,
                lookup.tls_handshakes,
                lookup.network_events,
                netinfodb,
                ip_to_domain,
            ),
        )


def base_processor(
    msmt: BaseMeasurement,
    db: DatabaseConnection,
    netinfodb: NetinfoDB,
) -> None:
    write_observations_to_db(db, [NettestObservation.from_measurement(msmt, netinfodb)])


def domains_in_a_day(
    day: date, db: ClickhouseConnection, probe_cc: Optional[str]
) -> List[str]:
    q = """SELECT DISTINCT(domain_name) FROM obs_dns
    WHERE timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    """
    params = one_day_dict(day)
    if probe_cc:
        q += "AND probe_cc = %(probe_cc)s"
        params["probe_cc"] = probe_cc
    res = db.execute(q, params)
    assert isinstance(res, list)
    return [row[0] for row in res]


def dns_observations_by_session(
    day: date,
    domain_name: str,
    db: ClickhouseConnection,
    probe_cc: Optional[str] = None,
) -> Generator[List[DNSObservation], None, None]:
    # I wish I had an ORM...
    field_names = observation_field_names(DNSObservation)

    q_params = one_day_dict(day)
    q_params["domain_name"] = domain_name

    q = "SELECT "
    q += ",\n".join(field_names)
    q += """
    FROM obs_dns
    WHERE domain_name = %(domain_name)s
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    """
    if probe_cc:
        q += "AND probe_cc = %(probe_cc)s\n"
        q_params["probe_cc"] = probe_cc

    q += "ORDER BY report_id, measurement_uid;"

    # Put all the DNS observations from the same testing session into a list and yield it
    dns_obs_session = []
    last_obs_session_id = None
    res = db.execute(q, q_params)
    assert isinstance(res, list)
    for rows in res:
        obs_dict = {field_names[idx]: val for idx, val in enumerate(rows)}
        dns_obs = DNSObservation(**obs_dict)

        if last_obs_session_id and last_obs_session_id != dns_obs.report_id:
            yield dns_obs_session
            dns_obs_session = [dns_obs]
            last_obs_session_id = dns_obs.report_id
        else:
            dns_obs_session.append(dns_obs)
    if len(dns_obs_session) > 0:
        yield dns_obs_session


T = TypeVar("T", bound="Observation")


def observations_in_session(
    day: date,
    domain_name: str,
    obs_class: Type[T],
    report_id: str,
    db: ClickhouseConnection,
) -> List[T]:
    observation_list = []
    field_names = observation_field_names(obs_class)
    q = "SELECT "
    q += ",\n".join(field_names)
    q += " FROM " + obs_class.__table_name__
    q += """
    WHERE domain_name = %(domain_name)s
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    AND report_id = %(report_id)s;
    """
    q_params = one_day_dict(day)
    q_params["domain_name"] = domain_name
    q_params["report_id"] = report_id

    res = db.execute(q, q_params)
    assert isinstance(res, list)
    for rows in res:
        obs_dict = {field_names[idx]: val for idx, val in enumerate(rows)}
        observation_list.append(obs_class(**obs_dict))
    return observation_list


def websites_observation_group(
    day: date,
    domain_name: str,
    db: ClickhouseConnection,
    probe_cc: Optional[str] = None,
) -> Generator[
    Tuple[
        List[DNSObservation],
        List[TCPObservation],
        List[TLSObservation],
        List[HTTPObservation],
    ],
    None,
    None,
]:
    for dns_obs_list in dns_observations_by_session(day, domain_name, db, probe_cc):
        report_id = dns_obs_list[0].report_id
        tcp_o_list = observations_in_session(
            day,
            domain_name,
            TCPObservation,
            report_id,
            db,
        )
        tls_o_list = observations_in_session(
            day,
            domain_name,
            TLSObservation,
            report_id,
            db,
        )
        http_o_list = observations_in_session(
            day,
            domain_name,
            HTTPObservation,
            report_id,
            db,
        )

        # Drop all verdicts related to this session from the database.
        # XXX this should probably be refactored to be closer to the place where
        # we do the insert, but will require quite a bit of reorganizing of the
        # logic in here.
        db.execute(
            "ALTER TABLE verdict DELETE WHERE report_id = %(report_id)s",
            {"report_id": report_id},
        )
        yield dns_obs_list, tcp_o_list, tls_o_list, http_o_list


import certifi
import ssl
import socket


def is_tls_valid(ip, hostname):
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    context.load_verify_locations(certifi.where())

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0) as sock:
        sock.settimeout(1)
        with context.wrap_socket(sock, server_hostname=hostname) as conn:
            try:
                conn.connect((ip, 443))
            # TODO: do we care to distinguish these values?
            except ssl.SSLCertVerificationError:
                return False
            except ssl.SSLError:
                return False
            except socket.timeout:
                return False
            except socket.error:
                return False
            except:
                return False
    return True


def get_extra_dns_consistency_tls_baseline(
    dns_baseline: DNSBaseline,
    dns_o_list: List[DNSObservation],
    db: ClickhouseConnection,
) -> List[str]:
    domain_name = dns_o_list[0].domain_name
    missing_answers = set(map(lambda a: a.answer, dns_o_list))
    for a in list(missing_answers):
        if a is None:
            missing_answers.discard(a)
            continue

        try:
            if is_ip_bogon(a):
                missing_answers.discard(a)
        except ValueError:
            missing_answers.discard(a)

    for a in dns_baseline.tls_consistent_answers:
        missing_answers.discard(a)

    new_tls_consistent_ips = []
    res = db.execute(
        "SELECT DISTINCT ip FROM dns_consistency_tls_baseline WHERE ip IN %(ip_list)s AND domain_name = %(domain_name)s",
        {"ip_list": list(missing_answers), "domain_name": domain_name},
    )
    assert isinstance(res, list)
    for row in res:
        missing_answers.discard(row[0])
        new_tls_consistent_ips.append(row[0])

    rows_to_write = []
    for ip in missing_answers:
        timestamp = datetime.now()
        if is_tls_valid(ip, domain_name):
            rows_to_write.append((ip, domain_name, timestamp))
            new_tls_consistent_ips.append(ip)

    if len(rows_to_write) > 0:
        db.execute(
            "INSERT INTO dns_consistency_tls_baseline (ip, domain_name, timestamp) VALUES",
            rows_to_write,
        )
    return new_tls_consistent_ips


def generate_website_verdicts(
    day: date,
    db: ClickhouseConnection,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
    probe_cc: Optional[str] = None,
):
    with logging_redirect_tqdm():
        for domain_name in tqdm(domains_in_a_day(day, db, probe_cc)):
            log.debug(f"Generating verdicts for {domain_name}")
            dns_baseline = make_dns_baseline(day, domain_name, db)
            http_baseline_map = make_http_baseline_map(day, domain_name, db)
            tcp_baseline_map = make_tcp_baseline_map(day, domain_name, db)

            for (
                dns_o_list,
                tcp_o_list,
                tls_o_list,
                http_o_list,
            ) in websites_observation_group(day, domain_name, db, probe_cc=probe_cc):
                extra_tls_consistent_answers = get_extra_dns_consistency_tls_baseline(
                    dns_baseline, dns_o_list, db
                )
                dns_baseline.tls_consistent_answers = list(
                    set(
                        list(dns_baseline.tls_consistent_answers)
                        + extra_tls_consistent_answers
                    )
                )
                yield from make_website_verdicts(
                    dns_o_list,
                    dns_baseline,
                    fingerprintdb,
                    netinfodb,
                    tcp_o_list,
                    tcp_baseline_map,
                    tls_o_list,
                    http_o_list,
                    http_baseline_map,
                )


verdict_generators = [generate_website_verdicts]

nettest_processors = {
    "web_connectivity": web_connectivity_processor,
    "dnscheck": dnscheck_processor,
    "tor": tor_processor,
}


def process_day(
    db: Union[ClickhouseConnection, CSVConnection],
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
    day: date,
    test_name=[],
    probe_cc=[],
    start_at_idx=0,
    skip_verdicts=False,
    fast_fail=False,
) -> Tuple[float, date]:
    t0 = time.monotonic()
    with tqdm(unit="B", unit_scale=True) as pbar:

        def progress_callback(p: MeasurementListProgress):
            if p.progress_status == ProgressStatus.LISTING:
                if not pbar.total:
                    pbar.total = p.total_prefixes
                pbar.update(1)
                pbar.set_description(
                    f"listed {p.total_file_entries} files in {p.current_prefix_idx}/{p.total_prefixes} prefixes"
                )
                return

            if p.progress_status == ProgressStatus.DOWNLOAD_BEGIN:
                pbar.unit = "B"
                pbar.reset(total=p.total_file_entry_bytes)

            pbar.set_description(
                f"downloading {p.current_file_entry_idx}/{p.total_file_entries} files"
            )
            pbar.update(p.current_file_entry_bytes)

        for idx, msmt_dict in enumerate(
            iter_measurements(
                probe_cc=probe_cc,
                test_name=test_name,
                start_day=day,
                end_day=day + timedelta(days=1),
                progress_callback=progress_callback,
            )
        ):
            pbar.set_description(f"idx {idx}")
            if idx < start_at_idx:
                continue
            try:
                msmt = load_measurement(msmt_dict)
                base_processor(msmt, db, netinfodb)
                processor = nettest_processors.get(msmt.test_name, default_processor)
                processor(
                    msmt,
                    db,
                    fingerprintdb,
                    netinfodb,
                )
            except Exception as exc:
                log.error(f"failed at idx:{idx} {exc}")
                with open("bad_msmts.jsonl", "ab+") as out_file:
                    out_file.write(orjson.dumps(msmt_dict))
                    out_file.write(b"\n")
                with open("bad_msmts_fail_log.txt", "a+") as out_file:
                    out_file.write(traceback.format_exc())
                    out_file.write("ENDTB----\n")
                if fast_fail:
                    raise exc

    if isinstance(db, ClickhouseConnection) and not skip_verdicts:
        write_verdicts_to_db(
            db,
            generate_website_verdicts(
                day,
                db,
                fingerprintdb,
                netinfodb,
            ),
        )

    return time.monotonic() - t0, day



from collections import defaultdict
from functools import partial
import gzip
import logging
import multiprocessing
from typing import List

import orjson
from tqdm.contrib.logging import tqdm_logging_redirect
from oonidata.dataclient import FileEntry, get_file_entries, iter_measurements
from oonidata.datautils import trim_measurement


log = logging.getLogger("oonidata")
logging.basicConfig(level=logging.INFO)


def make_filename(args, fe):
    flags = ""
    if args.max_string_size:
        flags = f"_max{args.max_string_size}"
    ts = fe.timestamp.strftime("%Y%m%d%H")
    filename = f"{ts}_{fe.probe_cc}_{fe.testname}{flags}.jsonl.gz"
    return filename


def download_file_entry_list(fe_list: List[FileEntry], args):
    """
    Download a list of file entries to the output dir.

    It assumes that the list of file entries in the list are all pertinent to
    the same testname, probe_cc, hour tuple
    """
    total_fe_size = 0
    output_dir = (
        args.output_dir
        / fe_list[0].testname
        / fe_list[0].timestamp.strftime("%Y-%m-%d")
    )
    output_dir.mkdir(parents=True, exist_ok=True)

    output_path = output_dir / make_filename(args, fe_list[0])

    with gzip.open(output_path.with_suffix(".tmp"), "wb") as out_file:
        for fe in fe_list:
            assert fe.testname == fe_list[0].testname
            assert fe.timestamp == fe_list[0].timestamp
            assert fe.probe_cc == fe_list[0].probe_cc
            total_fe_size += fe.size

        for msmt_dict in iter_measurements(
            start_day=args.start_day,
            end_day=args.end_day,
            probe_cc=args.probe_cc,
            test_name=args.test_name,
            file_entries=fe_list,
        ):
            if args.max_string_size:
                msmt_dict = trim_measurement(msmt_dict, args.max_string_size)
            out_file.write(orjson.dumps(msmt_dict))
            out_file.write(b"\n")

    output_path.with_suffix(".tmp").rename(output_path)
    return total_fe_size


def run_sync(args):
    log.info(
        f"Downloading measurement in s3 for {args.start_day} - {args.end_day} probe_cc:"
        f" {args.probe_cc}"
    )
    testnames = set()
    if args.test_name is not None:
        testnames = set(list(map(lambda x: x.lower().replace("_", ""), args.test_name)))

    ccs = set()
    if args.probe_cc is not None:
        ccs = set(list(map(lambda x: x.upper(), args.probe_cc)))

    with tqdm_logging_redirect(unit_scale=True) as pbar:

        def cb_list_fe(p):
            if p.current_prefix_idx == 0:
                pbar.total = p.total_prefixes
                pbar.update(0)
                pbar.set_description("starting prefix listing")
                return
            pbar.update(1)
            pbar.set_description(
                f"listed {p.total_file_entries} files in {p.current_prefix_idx}/{p.total_prefixes} prefixes"
            )

        log.info("Listing file entries...")
        all_file_entries = get_file_entries(
            start_day=args.start_day,
            end_day=args.end_day,
            ccs=ccs,
            testnames=testnames,
            from_cans=True,
            progress_callback=cb_list_fe,
        )

        total_fe_size = 0
        to_download_fe = defaultdict(list)
        for fe in all_file_entries:
            if (
                args.output_dir
                / fe.testname
                / fe.timestamp.strftime("%Y%-m-%d")
                / make_filename(args, fe)
            ).exists():
                continue

            ts = fe.timestamp.strftime("%Y%m%d%H")
            # We group based on this key, so each process is writing to the same file.
            # Each raw folder can have multiple files on a given hour
            key = f"{ts}-{fe.testname}-{fe.probe_cc}"
            to_download_fe[key].append(fe)
            total_fe_size += fe.size

        pbar.unit = "B"
        pbar.reset(total=total_fe_size)
        pbar.set_description("downloading files")
        download_count = 0
        with multiprocessing.Pool() as pool:
            for fe_size in pool.imap_unordered(
                partial(
                    download_file_entry_list,
                    args=args,
                ),
                to_download_fe.values(),
            ):
                download_count += 1
                pbar.update(fe_size)
                pbar.set_description(
                    f"downloaded {download_count}/{len(to_download_fe)}"
                )



import ipaddress
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, List, Tuple, Generator, Any, Mapping, Dict
from datetime import datetime, date

from urllib.parse import urlparse

from oonidata.fingerprintdb import FingerprintDB
from oonidata.netinfo import NetinfoDB

from oonidata.datautils import one_day_dict

from oonidata.observations import (
    DNSObservation,
    HTTPObservation,
    Observation,
    TCPObservation,
    TLSObservation,
)
from oonidata.dataformat import WebConnectivityControl, WebConnectivity
from oonidata.db.connections import ClickhouseConnection

import logging

log = logging.getLogger("oonidata.processing")


class Outcome(Enum):
    # k: everything is OK
    OK = "k"
    # b: blocking is happening with an unknown scope
    BLOCKED = "b"
    # n: national level blocking
    NATIONAL_BLOCK = "n"
    # i: isp level blocking
    ISP_BLOCK = "i"
    # l: local blocking (school, office, home network)
    LOCAL_BLOCK = "l"
    # s: server-side blocking
    SERVER_SIDE_BLOCK = "s"
    # d: the subject is down
    DOWN = "d"
    # t: this is a signal indicating some form of network throttling
    THROTTLING = "t"


def fp_scope_to_outcome(scope: Optional[str]) -> Outcome:
    # "nat" national level blockpage
    # "isp" ISP level blockpage
    # "prod" text pattern related to a middlebox product
    # "inst" text pattern related to a voluntary instition blockpage (school, office)
    # "vbw" vague blocking word
    # "fp" fingerprint for false positives
    if scope == "nat":
        return Outcome.NATIONAL_BLOCK
    elif scope == "isp":
        return Outcome.ISP_BLOCK
    elif scope == "inst":
        return Outcome.LOCAL_BLOCK
    elif scope == "fp":
        return Outcome.SERVER_SIDE_BLOCK
    return Outcome.BLOCKED


@dataclass
class Verdict:
    measurement_uid: str
    observation_id: str
    report_id: str
    input: str
    timestamp: datetime

    probe_asn: int
    probe_cc: str

    probe_as_org_name: str
    probe_as_cc: str

    network_type: str

    resolver_ip: Optional[str]
    resolver_asn: Optional[int]
    resolver_as_org_name: Optional[str]
    resolver_as_cc: Optional[str]
    resolver_cc: Optional[str]

    confidence: float

    subject: str
    subject_category: str
    subject_detail: str

    outcome: Outcome

    # This will include a more detailed breakdown of the outcome, for example it
    # can be "dns.nxdomain"
    outcome_detail: str


def make_verdict_from_obs(
    obs: Observation,
    confidence: float,
    subject: str,
    subject_category: str,
    subject_detail: str,
    outcome: Outcome,
    outcome_detail: str,
) -> Verdict:
    return Verdict(
        measurement_uid=obs.measurement_uid,
        observation_id=obs.observation_id,
        report_id=obs.report_id,
        input=obs.input,
        timestamp=obs.timestamp,
        probe_asn=obs.probe_asn,
        probe_cc=obs.probe_cc,
        probe_as_org_name=obs.probe_as_org_name,
        probe_as_cc=obs.probe_as_cc,
        network_type=obs.network_type,
        resolver_ip=obs.resolver_ip,
        resolver_asn=obs.resolver_asn,
        resolver_as_org_name=obs.resolver_as_org_name,
        resolver_as_cc=obs.resolver_as_cc,
        resolver_cc=obs.resolver_cc,
        confidence=confidence,
        subject=subject,
        subject_category=subject_category,
        subject_detail=subject_detail,
        outcome=outcome,
        outcome_detail=outcome_detail,
    )


@dataclass
class TCPBaseline:
    address: str
    reachable_cc_asn: List[Tuple[str, int]] = field(default_factory=list)
    unreachable_cc_asn: List[Tuple[str, int]] = field(default_factory=list)


def make_tcp_baseline_map(
    day: date, domain_name: str, db: ClickhouseConnection
) -> Mapping[str, TCPBaseline]:
    tcp_baseline_map = {}
    q_params = one_day_dict(day)
    q_params["domain_name"] = domain_name

    q = """SELECT probe_cc, probe_asn, ip, port, failure FROM obs_tcp
    WHERE domain_name = %(domain_name)s 
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    GROUP BY probe_cc, probe_asn, ip, port, failure;
    """
    res = db.execute(q, q_params)
    if isinstance(res, list) and len(res) > 0:
        for probe_cc, probe_asn, ip, port, failure in res:
            address = f"{ip}:{port}"
            tcp_baseline_map[address] = tcp_baseline_map.get(
                address, TCPBaseline(address)
            )
            if not failure:
                tcp_baseline_map[address].reachable_cc_asn.append((probe_cc, probe_asn))
            else:
                tcp_baseline_map[address].unreachable_cc_asn.append(
                    (probe_cc, probe_asn)
                )
    return tcp_baseline_map


@dataclass
class HTTPBaseline:
    url: str
    failure_cc_asn: List[Tuple[str, int]] = field(default_factory=list)
    ok_cc_asn: List[Tuple[str, int]] = field(default_factory=list)

    response_body_length: int = 0
    response_body_sha1: str = ""
    response_body_title: str = ""
    response_body_meta_title: str = ""

    response_status_code: int = 0


def maybe_get_first(l: list, default_value: Any = None) -> Optional[Any]:
    try:
        return l[0]
    except IndexError:
        return default_value


def make_http_baseline_map(
    day: date, domain_name: str, db: ClickhouseConnection
) -> Mapping[str, HTTPBaseline]:
    http_baseline_map = {}

    q_params = one_day_dict(day)
    q_params["domain_name"] = domain_name

    q = """SELECT probe_cc, probe_asn, request_url, failure FROM obs_http
    WHERE domain_name = %(domain_name)s 
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    GROUP BY probe_cc, probe_asn, request_url, failure;
    """
    res = db.execute(q, q_params)
    if isinstance(res, list) and len(res) > 0:
        for probe_cc, probe_asn, request_url, failure in res:
            http_baseline_map[request_url] = http_baseline_map.get(
                request_url, HTTPBaseline(request_url)
            )
            if not failure:
                http_baseline_map[request_url].ok_cc_asn.append((probe_cc, probe_asn))
            else:
                http_baseline_map[request_url].failure_cc_asn.append(
                    (probe_cc, probe_asn)
                )

    q = """SELECT request_url,
    topK(1)(response_body_sha1),
    topK(1)(response_body_length),
    topK(1)(response_body_title),
    topK(1)(response_body_meta_title),
    topK(1)(response_status_code)
    FROM obs_http
    WHERE failure IS NULL
    AND domain_name = %(domain_name)s 
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    GROUP BY request_url;
    """
    res = db.execute(q, q_params)
    if isinstance(res, list) and len(res) > 0:
        for (
            request_url,
            response_body_sha1,
            response_body_length,
            response_body_title,
            response_body_meta_title,
            response_status_code,
        ) in res:
            http_baseline_map[request_url] = http_baseline_map.get(
                request_url, HTTPBaseline(request_url)
            )
            http_baseline_map[request_url].response_body_sha1 = maybe_get_first(
                response_body_sha1, ""
            )
            http_baseline_map[request_url].response_body_length = maybe_get_first(
                response_body_length, ""
            )
            http_baseline_map[request_url].response_body_title = maybe_get_first(
                response_body_title, ""
            )
            http_baseline_map[request_url].response_body_meta_title = maybe_get_first(
                response_body_meta_title, ""
            )
            http_baseline_map[request_url].response_status_code = maybe_get_first(
                response_status_code, ""
            )

    return http_baseline_map


@dataclass
class DNSBaseline:
    domain: str
    nxdomain_cc_asn: List[Tuple[str, int]] = field(default_factory=list)
    failure_cc_asn: List[Tuple[str, int]] = field(default_factory=list)
    ok_cc_asn: List[Tuple[str, int]] = field(default_factory=list)
    tls_consistent_answers: List[str] = field(default_factory=list)
    answers_map: Dict[str, List[Tuple[str, str]]] = field(default_factory=dict)


def make_dns_baseline(
    day: date, domain_name: str, db: ClickhouseConnection
) -> DNSBaseline:
    dns_baseline = DNSBaseline(domain_name)

    q_params = one_day_dict(day)
    q_params["domain_name"] = domain_name

    q = """SELECT DISTINCT(ip) FROM obs_tls
    WHERE is_certificate_valid = 1 
    AND domain_name = %(domain_name)s 
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s;
    """
    res = db.execute(q, q_params)
    if isinstance(res, list) and len(res) > 0:
        dns_baseline.tls_consistent_answers = [row[0] for row in res]

    q = """SELECT probe_cc, probe_asn, failure, answer FROM obs_dns
    WHERE domain_name = %(domain_name)s 
    AND timestamp >= %(start_day)s
    AND timestamp <= %(end_day)s
    GROUP BY probe_cc, probe_asn, failure, answer;
    """
    res = db.execute(q, q_params)
    if isinstance(res, list) and len(res) > 0:
        for probe_cc, probe_asn, failure, ip in res:
            if not failure:
                dns_baseline.ok_cc_asn.append((probe_cc, probe_asn))
                dns_baseline.answers_map[probe_cc] = dns_baseline.answers_map.get(
                    probe_cc, []
                )
                if ip:
                    dns_baseline.answers_map[probe_cc].append((probe_asn, ip))
                else:
                    log.error(
                        f"No IP present for {domain_name} {probe_cc} ({probe_asn}) in baseline"
                    )
            else:
                dns_baseline.failure_cc_asn.append((probe_cc, probe_asn))
                if failure == "dns_nxdomain_error":
                    dns_baseline.nxdomain_cc_asn.append((probe_cc, probe_asn))

    return dns_baseline


def is_dns_consistent(
    dns_o: DNSObservation, dns_b: DNSBaseline, netinfodb: NetinfoDB
) -> Tuple[bool, float]:
    if not dns_o.answer:
        return False, 0

    try:
        ipaddress.ip_address(dns_o.answer)
    except ValueError:
        # Not an IP, er can't do much to validate it
        return False, 0

    if dns_o.answer in dns_b.tls_consistent_answers:
        return True, 1.0

    baseline_asns = set()
    baseline_as_org_names = set()

    for ip in dns_b.tls_consistent_answers:
        ip_info = netinfodb.lookup_ip(dns_o.timestamp, ip)
        if ip_info:
            baseline_asns.add(ip_info.as_info.asn)
            baseline_as_org_names.add(ip_info.as_info.as_org_name.lower())

    if dns_o.answer_asn in baseline_asns:
        return True, 0.9

    # XXX maybe with the org_name we can also do something like levenshtein
    # distance to get more similarities
    if (
        dns_o.answer_as_org_name
        and dns_o.answer_as_org_name.lower() in baseline_as_org_names
    ):
        return True, 0.9

    other_answers = dns_b.answers_map.copy()
    other_answers.pop(dns_o.probe_cc, None)
    other_ips = {}
    other_asns = {}
    for answer_list in other_answers.values():
        for _, ip in answer_list:

            other_ips[ip] = other_ips.get(ip, 0)
            other_ips[ip] += 1
            if ip is None:
                log.error(f"Missing ip for {dns_o.domain_name}")
                continue
            ip_info = netinfodb.lookup_ip(dns_o.timestamp, ip)
            if ip_info:
                asn = ip_info.as_info.asn
                other_asns[asn] = other_asns.get(ip, 0)
                other_asns[asn] += 1

    if dns_o.answer in other_ips:
        x = other_ips[dns_o.answer]
        # This function was derived by looking for an exponential function in
        # the form f(x) = c1*a^x + c2 and solving for f(0) = 0 and f(10) = 1,
        # giving us a function in the form f(x) = (a^x - 1) / (a^10 - 1). We
        # then choose the magic value of 0.6 by looking for a solution in a
        # where f(1) ~= 0.5, doing a bit of plots and choosing a curve that
        # looks reasonably sloped.
        y = (pow(0.5, x) - 1) / (pow(0.5, 10) - 1)
        return True, min(0.9, 0.8 * y)

    if dns_o.answer in other_asns:
        x = other_asns[dns_o.answer_asn]
        y = (pow(0.5, x) - 1) / (pow(0.5, 10) - 1)
        return True, min(0.8, 0.7 * y)

    x = len(baseline_asns)
    y = (pow(0.5, x) - 1) / (pow(0.5, 10) - 1)
    return False, min(0.9, 0.8 * y)


def make_website_tcp_verdicts(
    tcp_o: TCPObservation, tcp_b: TCPBaseline
) -> Optional[Verdict]:
    outcome = Outcome.OK
    confidence = 1
    outcome_detail = ""

    if tcp_o.failure:
        unreachable_cc_asn = list(tcp_b.unreachable_cc_asn)
        try:
            unreachable_cc_asn.remove((tcp_o.probe_cc, tcp_o.probe_asn))
        except ValueError:
            log.info(
                "missing failure in tcp baseline. You are probably using a control derived baseline."
            )

        reachable_count = len(tcp_b.reachable_cc_asn)
        unreachable_count = len(unreachable_cc_asn)
        if reachable_count > unreachable_count:
            # We are adding back 1 because we removed it above and it avoid a divide by zero
            confidence = reachable_count / (reachable_count + unreachable_count + 1)
            outcome = Outcome.BLOCKED
        elif unreachable_count > reachable_count:
            confidence = (unreachable_count + 1) / (
                reachable_count + unreachable_count + 1
            )
            outcome = Outcome.DOWN

        outcome_detail = f"tcp.{tcp_o.failure}"

    if outcome != Outcome.OK:
        return make_verdict_from_obs(
            tcp_o,
            confidence=confidence,
            subject=tcp_o.domain_name,
            subject_detail=f"{tcp_o.ip}:{tcp_o.port}",
            subject_category="website",
            outcome=outcome,
            outcome_detail=outcome_detail,
        )


def make_website_dns_verdict(
    dns_o: DNSObservation,
    dns_b: DNSBaseline,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
) -> Optional[Verdict]:
    if dns_o.fingerprint_id:
        fp = fingerprintdb.get_fp(dns_o.fingerprint_id)
        outcome = fp_scope_to_outcome(fp.scope)
        confidence = 1.0
        # If we see the fingerprint in an unexpected country we should
        # significantly reduce the confidence in the block
        if (
            dns_o.probe_cc
            and fp.expected_countries
            and len(fp.expected_countries) > 0
            and dns_o.probe_cc not in fp.expected_countries
        ):
            log.debug(
                f"Inconsistent probe_cc vs expected_countries {dns_o.probe_cc} != {fp.expected_countries}"
            )
            confidence = 0.7

        outcome_detail = "dns.blockpage"
        return make_verdict_from_obs(
            dns_o,
            confidence=confidence,
            subject=dns_o.domain_name,
            subject_detail=f"{dns_o.answer}",
            subject_category="website",
            outcome=Outcome.BLOCKED,
            outcome_detail=outcome_detail,
        )

    elif dns_o.answer_is_bogon and len(dns_b.tls_consistent_answers) > 0:
        outcome_detail = "dns.bogon"
        return make_verdict_from_obs(
            dns_o,
            confidence=0.9,
            subject=dns_o.domain_name,
            subject_detail=f"{dns_o.answer}",
            subject_category="website",
            outcome=Outcome.BLOCKED,
            outcome_detail="dns.bogon",
        )

    elif dns_o.failure:
        failure_cc_asn = list(dns_b.failure_cc_asn)
        try:
            failure_cc_asn.remove((dns_o.probe_cc, dns_o.probe_asn))
        except ValueError:
            log.info(
                "missing failure for the probe in the baseline. You are probably using a control derived baseline."
            )

        failure_count = len(failure_cc_asn)
        ok_count = len(dns_b.ok_cc_asn)

        if dns_o.failure == "dns_nxdomain_error":
            nxdomain_cc_asn = list(dns_b.nxdomain_cc_asn)
            try:
                nxdomain_cc_asn.remove((dns_o.probe_cc, dns_o.probe_asn))
            except ValueError:
                log.info(
                    "missing nx_domain failure for the probe in the baseline. You are probably using a control derived baseline."
                )

            nxdomain_count = len(nxdomain_cc_asn)
            if ok_count > nxdomain_count:
                # We give a bit extra weight to an NXDOMAIN compared to other failures
                confidence = ok_count / (ok_count + nxdomain_count + 1)
                confidence = min(0.8, confidence * 1.5)
                outcome = Outcome.BLOCKED
                outcome_detail = "dns.nxdomain"
            else:
                confidence = (nxdomain_count + 1) / (ok_count + nxdomain_count + 1)
                outcome = Outcome.DOWN
                outcome_detail = "dns.nxdomain"
        elif ok_count > failure_count:
            confidence = ok_count / (ok_count + failure_count + 1)
            outcome = Outcome.BLOCKED
            outcome_detail = f"dns.{dns_o.failure}"
        else:
            confidence = (failure_count + 1) / (ok_count + failure_count + 1)
            outcome = Outcome.DOWN
            outcome_detail = f"dns.{dns_o.failure}"
        return make_verdict_from_obs(
            dns_o,
            confidence=confidence,
            subject=dns_o.domain_name,
            subject_detail=f"{dns_o.answer}",
            subject_category="website",
            outcome=outcome,
            outcome_detail=outcome_detail,
        )

    elif dns_o.is_tls_consistent == False:
        outcome_detail = "dns.inconsistent.tls_mismatch"
        return make_verdict_from_obs(
            dns_o,
            confidence=0.8,
            subject=dns_o.domain_name,
            subject_detail=f"{dns_o.answer}",
            subject_category="website",
            outcome=Outcome.BLOCKED,
            outcome_detail=outcome_detail,
        )

    elif dns_o.is_tls_consistent == None:
        # If we are in this case, it means we weren't able to determine the
        # consistency of the DNS query using TLS. This is the case either
        # because the tested site is not in HTTPS and therefore we didn't
        # generate a TLS measurement for it or because the target IP isn't
        # listening on HTTPS (which is quite fishy).
        # In either case we should flag these with being somewhat likely to be
        # blocked.
        ip_based_consistency, consistency_confidence = is_dns_consistent(
            dns_o, dns_b, netinfodb
        )
        if ip_based_consistency is False and consistency_confidence > 0:
            confidence = consistency_confidence
            outcome_detail = "dns.inconsistent.generic"
            # If the answer ASN is the same as the probe_asn, it's more likely
            # to be a blockpage
            if dns_o.answer_asn == dns_o.probe_asn:
                outcome_detail = "dns.inconsistent.asn_match"
                confidence = 0.8
            # same for the answer_cc
            elif dns_o.answer_as_cc == dns_o.probe_cc:
                outcome_detail = "dns.inconsistent.cc_match"
                confidence = 0.7
            return make_verdict_from_obs(
                dns_o,
                confidence=confidence,
                subject=dns_o.domain_name,
                subject_detail=f"{dns_o.answer}",
                subject_category="website",
                outcome=Outcome.BLOCKED,
                outcome_detail=outcome_detail,
            )
    # No blocking detected
    return None


def make_website_tls_verdict(
    tls_o: TLSObservation, prev_verdicts: List[Verdict]
) -> Optional[Verdict]:
    if tls_o.is_certificate_valid == False:
        # We only consider it to be a TLS level verdict in cases when there is a
        # certificate mismatch, but there was no DNS inconsistency.
        # If the DNS was inconsistent, we will just count the DNS verdict
        if (
            len(
                list(
                    filter(
                        lambda v: v.outcome_detail.startswith("dns.")
                        and v.subject_detail == tls_o.ip,
                        prev_verdicts,
                    )
                )
            )
            > 0
        ):
            return

        # TODO: this is wrong. We need to consider the baseline to establish TLS
        # MITM, because the cert might be invalid also from other location (eg.
        # it expired) and not due to censorship.
        outcome_detail = "tls.mitm"
        return make_verdict_from_obs(
            tls_o,
            confidence=1,
            subject=tls_o.domain_name,
            subject_detail=f"{tls_o.ip}:{tls_o.port}",
            subject_category="website",
            outcome=Outcome.BLOCKED,
            outcome_detail=outcome_detail,
        )
    elif tls_o.failure:
        if (
            len(
                list(
                    filter(
                        lambda v: v.outcome_detail.startswith("tcp.")
                        and v.subject_detail == f"{tls_o.ip}:443",
                        prev_verdicts,
                    )
                )
            )
            > 0
        ):
            return

        # We only consider it to be a TLS level verdict if we haven't seen any
        # blocks in TCP
        outcome_detail = f"tls.{tls_o.failure}"
        confidence = 0.5

        if tls_o.tls_handshake_read_count == 0 and tls_o.tls_handshake_write_count == 1:
            # This means we just wrote the TLS ClientHello, let's give it a bit
            # more confidence in it being a block
            confidence = 0.7

        if tls_o.failure in ("connection_closed", "connection_reset"):
            confidence += 0.2

        return make_verdict_from_obs(
            tls_o,
            confidence=confidence,
            subject=tls_o.domain_name,
            subject_detail=f"{tls_o.ip}:{tls_o.port}",
            subject_category="website",
            outcome=Outcome.BLOCKED,
            outcome_detail=outcome_detail,
        )


def make_website_http_verdict(
    http_o: HTTPObservation,
    http_b: HTTPBaseline,
    prev_verdicts: List[Verdict],
    fingerprintdb: FingerprintDB,
) -> Optional[Verdict]:
    if http_o.failure:
        # For HTTP requests we ignore cases in which we detected the blocking
        # already to be happening via DNS or TCP.
        if not http_o.request_is_encrypted and (
            len(
                list(
                    filter(
                        lambda v: v.outcome_detail.startswith("dns.")
                        or (
                            v.outcome_detail.startswith("tcp.")
                            and v.subject_detail.endswith(":80")
                        ),
                        prev_verdicts,
                    )
                )
            )
            > 0
        ):
            return

        # Similarly for HTTPS we ignore cases when the block is done via TLS or TCP
        if http_o.request_is_encrypted and (
            len(
                list(
                    filter(
                        lambda v: v.outcome_detail.startswith("dns.")
                        or (
                            v.outcome_detail.startswith("tcp.")
                            and v.subject_detail.endswith(":443")
                        )
                        or v.outcome_detail.startswith("tls."),
                        prev_verdicts,
                    )
                )
            )
            > 0
        ):
            return

        failure_cc_asn = list(http_b.failure_cc_asn)
        try:
            failure_cc_asn.remove((http_o.probe_cc, http_o.probe_asn))
        except ValueError:
            log.info(
                "missing failure in http baseline. Either something is wrong or you are using a control derived baseline"
            )

        failure_count = len(failure_cc_asn)
        ok_count = len(http_b.ok_cc_asn)
        if ok_count > failure_count:
            # We are adding back 1 because we removed it above and it avoid a divide by zero
            confidence = ok_count / (ok_count + failure_count + 1)
            outcome = Outcome.BLOCKED
        else:
            confidence = (failure_count + 1) / (ok_count + failure_count + 1)
            outcome = Outcome.DOWN

        outcome_detail = "http."
        if http_o.request_is_encrypted:
            outcome_detail = "https."
        outcome_detail += http_o.failure
        return make_verdict_from_obs(
            http_o,
            confidence=confidence,
            subject=http_o.domain_name,
            subject_detail="",
            subject_category="website",
            outcome=outcome,
            outcome_detail=outcome_detail,
        )
    elif http_o.response_matches_blockpage:
        outcome = Outcome.BLOCKED
        confidence = 0.7
        if http_o.request_is_encrypted:
            confidence = 0
        elif http_o.fingerprint_country_consistent:
            confidence = 1

        for fp_name in http_o.response_fingerprints:
            fp = fingerprintdb.get_fp(fp_name)
            if fp.scope:
                outcome = fp_scope_to_outcome(fp.scope)
                break

        return make_verdict_from_obs(
            http_o,
            confidence=confidence,
            subject=http_o.domain_name,
            subject_detail="",
            subject_category="website",
            outcome=outcome,
            outcome_detail="http.blockpage",
        )

    elif not http_o.request_is_encrypted:
        if http_o.response_matches_false_positive:
            return
        if http_o.response_body_title == http_b.response_body_title:
            return
        if http_o.response_body_meta_title == http_b.response_body_meta_title:
            return
        if http_o.response_body_sha1 == http_b.response_body_sha1:
            return

        if (
            http_o.response_body_length
            and http_b.response_body_length
            and (
                (http_o.response_body_length + 1.0)
                / (http_b.response_body_length + 1.0)
                < 0.7
            )
        ):
            return make_verdict_from_obs(
                http_o,
                confidence=0.6,
                subject=http_o.domain_name,
                subject_detail="",
                subject_category="website",
                outcome=Outcome.BLOCKED,
                outcome_detail="http.bodydiff",
            )


def make_dns_baseline_from_control(
    msmt_input: str, control: WebConnectivityControl
) -> DNSBaseline:
    domain_name = urlparse(msmt_input).hostname

    assert domain_name is not None, "domain_name is None"

    if not control or not control.dns:
        return DNSBaseline(domain=domain_name)

    nxdomain_cc_asn = []
    if control.dns.failure == "dns_nxdomain_error":
        nxdomain_cc_asn.append(("ZZ", 0))

    ok_cc_asn = []
    failure_cc_asn = []
    if control.dns.failure is not None:
        failure_cc_asn.append(("ZZ", 0))
    else:
        ok_cc_asn.append(("ZZ", 0))

    answers_map = {}
    if control.dns.addrs:
        answers_map["ZZ"] = [(0, ip) for ip in control.dns.addrs]

    return DNSBaseline(
        domain=domain_name,
        answers_map=answers_map,
        ok_cc_asn=ok_cc_asn,
        nxdomain_cc_asn=nxdomain_cc_asn,
        failure_cc_asn=failure_cc_asn,
    )


def make_tcp_baseline_from_control(
    control: WebConnectivityControl,
) -> Mapping[str, TCPBaseline]:
    if not control or not control.tcp_connect:
        return {}

    tcp_b_map = {}
    for key, status in control.tcp_connect.items():
        if status.failure == None:
            tcp_b_map[key] = TCPBaseline(address=key, reachable_cc_asn=[("ZZ", 0)])
        else:
            tcp_b_map[key] = TCPBaseline(address=key, unreachable_cc_asn=[("ZZ", 0)])
    return tcp_b_map


def make_http_baseline_from_control(
    msmt: WebConnectivity, control: WebConnectivityControl
) -> Mapping[str, HTTPBaseline]:
    if not control or not control.http_request:
        return {}

    if not msmt.test_keys.requests:
        return {}

    http_b_map = {}
    # We make the baseline apply to every URL in the response chain, XXX evaluate how much this is a good idea
    for http_transaction in msmt.test_keys.requests:
        if not http_transaction.request:
            continue

        url = http_transaction.request.url
        if control.http_request.failure == None:
            http_b_map[url] = HTTPBaseline(
                url=url,
                response_body_title=control.http_request.title or "",
                response_body_length=control.http_request.body_length or 0,
                response_status_code=control.http_request.status_code or 0,
                response_body_meta_title="",
                response_body_sha1="",
                ok_cc_asn=[("ZZ", 0)],
            )
        else:
            http_b_map[url] = HTTPBaseline(
                url=url,
                response_body_title="",
                response_body_length=0,
                response_status_code=0,
                response_body_meta_title="",
                response_body_sha1="",
                failure_cc_asn=[("ZZ", 0)],
            )
    return http_b_map


def make_website_verdicts(
    dns_o_list: List[DNSObservation],
    dns_b: DNSBaseline,
    fingerprintdb: FingerprintDB,
    netinfodb: NetinfoDB,
    tcp_o_list: List[TCPObservation],
    tcp_b_map: Mapping[str, TCPBaseline],
    tls_o_list: List[TLSObservation],
    http_o_list: List[HTTPObservation],
    http_b_map: Mapping[str, HTTPBaseline],
) -> Generator[Verdict, None, List[str]]:
    """
    make_website_verdicts will yield many verdicts given some observations
    related to a website measurement.

    We MUST pass in DNS observations, but observations of other types are
    optional. This is to workaround the fact that not every version of OONI
    Probe was generating all types of observations.

    The order in which we compute the verdicts is important, since the knowledge
    of some form of blocking is relevant to being able to determine future
    methods of blocking.
    Examples of this include:
    * If you know that DNS is consistent and you see a TLS certificate
    validation error, you can conclude that it's a MITM
    * If you see that TCP connect is failing, you will not attribute a failing
    TLS to a TLS level interference (ex. SNI filtering)

    For some lists of observations we also need to pass in a baseline. The
    baseline is some groundtruth related to the targets being measured that are
    needed in order to draw some meaningful conclusion about it's blocking.
    We need this so that we are able to exclude instances in which the target is
    unavailable due to transient network failures.

    It is the job of who calls the make_website_verdicts function to build this
    baseline by either running queries against the database of observations or
    using some set of observations that are already in memory.

    The basic idea is that since we are going to be generating verdicts in
    windows of 24h, we would have to do the baselining only once for the 24h
    time window given a certain domain.
    """
    verdicts = []

    domain_name = dns_o_list[0].domain_name
    dns_verdicts = []
    for dns_o in dns_o_list:
        assert (
            domain_name == dns_o.domain_name
        ), f"Inconsistent domain_name in dns_o {dns_o.domain_name}"
        dns_v = make_website_dns_verdict(dns_o, dns_b, fingerprintdb, netinfodb)
        if dns_v:
            dns_verdicts.append(dns_v)
        else:
            # If we didn't get a DNS verdict from an observation, it means that
            # observation was a sign of everything being OK, hence we should
            # ignore all the previous DNS verdicts as likely false positives and
            # just consider no DNS level censorship to be happening.
            dns_verdicts = []
            break

    for dns_v in dns_verdicts:
        verdicts.append(dns_v)
        yield dns_v

    if tcp_o_list:
        for tcp_o in tcp_o_list:
            assert (
                domain_name == tcp_o.domain_name
            ), f"Inconsistent domain_name in tcp_o {tcp_o.domain_name}"
            tcp_b = tcp_b_map.get(f"{tcp_o.ip}:{tcp_o.port}")
            tcp_v = make_website_tcp_verdicts(tcp_o, tcp_b) if tcp_b else None
            if tcp_v:
                verdicts.append(tcp_v)
                yield tcp_v

    if tls_o_list:
        for tls_o in tls_o_list:
            assert (
                domain_name == tls_o.domain_name
            ), f"Inconsistent domain_name in tls_o {tls_o.domain_name}"
            tls_v = make_website_tls_verdict(tls_o, verdicts)
            if tls_v:
                verdicts.append(tls_v)
                yield tls_v

    if http_o_list:
        for http_o in http_o_list:
            assert (
                domain_name == http_o.domain_name
            ), f"Inconsistent domain_name in http_o {http_o.domain_name}"
            http_b = http_b_map.get(http_o.request_url)
            http_v = (
                make_website_http_verdict(http_o, http_b, verdicts, fingerprintdb)
                if http_b
                else None
            )
            if http_v:
                verdicts.append(http_v)
                yield http_v

    if len(verdicts) == 0:
        # We didn't generate any verdicts up to now, so it's reasonable to say
        # there is no interference happening for the given domain_name
        ok_verdict = make_verdict_from_obs(
            dns_o_list[0],
            confidence=0.9,
            subject=domain_name,
            subject_detail="",
            subject_category="website",
            outcome=Outcome.OK,
            outcome_detail="all",
        )
        yield ok_verdict
        verdicts.append(ok_verdict)
    return verdicts
